"""
/**********************************************************************
FILE     : $RCSfile: taminotalk.py,v $
PURPOSE  : Client API for the Tamino XML database
NOTES    : Release version
AUTHOR   : Niels Mache
COPYRIGHT: (c) 2001, 2002 by struktur AG
DATE     : 11SEP2001
L.CHANGE : 04FEB2002
LICENSE  : see LICENSE.TXT
REVISION : $Revision: 1.17 $
VERSION  : $Id: taminotalk.py,v 1.17 2002/02/04 11:09:35 mache Exp $ (Author: $Author: mache $)

struktur AG            Phone: +49 711 896656 0
Junghansstr. 5         Fax:   +49 711 896656 10
70469 Stuttgart        email: info@struktur.de
GERMANY 

http://www.struktur.de
http://www.strukturag.com

**********************************************************************/

Client API for the Tamino XML database (version 2.3.1.1 and above).
The API implements basic functions for access and query of Tamino and transaction handling.
For convenience a function for XSLT formatting of XML output is provided.
This API contains fixes for bugs in Tamino 2.x. Beware of the following remaining bugs in Tamino
which are not fixed yet:

  - Tamino is unable to store Non-XML documents with length 0
    (*grin*).
  - Be care using the define method due to Tamino bugs in transaction handling. 


HISTORY:

11SEP2001:
  First release.

02JAN2002:
  Added commands: admin, deleteByID, diagnose
  Major changes: Changed return type of process and define. The commands now return an unparsed stream.  
  New commands: administration, diagnose


----------------------------------------------------------------------
  DISCLAIMER:

  THIS SOFTWARE IS PROVIDED BY STRUKTUR AG ``AS IS'' AND ANY
  EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL STRUKTUR AG OR ITS
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
----------------------------------------------------------------------
"""

__version__='$Revision: 1.17 $'[11:-2]



import types

# import DOM
import xml.dom
import xml.dom.ext
from xml.dom.ext.reader import PyExpat
from xml.dom.minidom import parse, parseString

# 4Suite XSLT
try:
    from xml.xslt.Processor import Processor
except:
    pass

# import SAX
from xml.sax import saxutils
from xml.sax import make_parser
from xml.sax.handler import feature_namespaces

# import urllib
import urllib, urllib2, socket, httplib, mimetools
from mimetypes import guess_type

# import HTTP1.1 multipart POST
from davclient import Resource 


class Results(saxutils.DefaultHandler):
    """
    Read the results of a Tamino request using SAX parser.
    """

    def __init__(self):
        self.sessionkey = None
        self.sessionid = None
        self.returnvalue = 0

    def getReturnValue(self):
        return self.returnvalue

    def getSession(self):
        return self.sessionid , self.sessionkey

    def hasSession(self):
        if self.sessionid is not None:
            return 1
        return 0

    def startDocument(self):
        self.sessionkey = None
        self.sessionid = None
        self.returnvalue = 0    # reset return value
    
    def startElement(self, name, attrs):
        if name == 'ino:response':
            self.sessionid = attrs.get('ino:sessionid', None)
            self.sessionkey = attrs.get('ino:sessionkey', None)

        #  XXX hack: get only the first return value. Other return values are ignored.
        if self.returnvalue == 0 and name == 'ino:message':
            self.returnvalue = int(attrs.get('ino:returnvalue', None))


class TaminoTalk:
    """
    This class connects to a Tamino XML database updates and queries XML documents.
    The database connection supports transactions, but it is not thread save now.
    """

    def __init__(self, server = None, database = None, collection = None, verbose = 0):
        self._reset()
        self._server = server
        self._database = database
        self._collection = collection
        self._verbose = verbose
        self._saxparser = make_parser()
        self._saxparser.setFeature(feature_namespaces, 0)
        self._results = Results()
        self._saxparser.setContentHandler(self._results)
        self._theStreamParser = self.getMiniDOMResultFromStream  # set minidom as the default parser 
        self._theStringParser = self.getMiniDOMResultFromString  # set minidom as the default parser 

    def _reset(self):
        self._url = None
        self._fstream = None
        self._xml_dom_object = None
        self._encoding = None
        self._parserType = 0
        self._sessionID = None
        self._sessionKey = None
        self._returnValue = 0


    def _del(self):
        if self._xml_dom_object is not None:
            self._xml_dom_object = None

            # delete the DOM object
            # not implemented
            #           if (self._parserType == 1):
            #               self._xml_dom_object.unlink()  # minidom
            #           if (self._parserType == 2):
            #               self._xml_dom_object = None  # expat

    def isVerbose(self, level):
        return self._verbose == level

    def setVerbose(self, verbose):
        self._verbose = verbose

    def connectTo(self, server, database = None, collection = None):
        """
        Set Tamino server URL, database and collection.
        Reset some variables.
        """
        self._reset()

        self._server = server
        self._database = database
        self._collection = collection

        self._url = 'http://' + self._server + '/tamino'
        if database is not None:
            self._url = self._url + '/' + self._database
        if collection is not None:
            self._url = self._url + '/' + collection


# --------------------------------------------------------------------------
# This code is subject to be changed:

    def getResultStream(self):
        """Get result stream."""
        return self._fstream

    def getDOM(self):
        """Get DOM object."""
        return self._xml_dom_object

    def getDOMResultFromStream(self, fstream):
        """Generate DOM level 2 from stream using expat."""
        reader = PyExpat.Reader()
        dom = reader.fromStream(fstream)
        return dom

    def getMiniDOMResultFromStream(self, fstream):
        """
        Generate DOM level 1 using minidom (DOM level 1) from stream. Approx. 5 times faster.
        """
        dom = parse(fstream)
        return dom

    def getDOMResultFromString(self, string):
        """
        Generate DOM level 2 from stream using expat.
        """
        reader = PyExpat.Reader()
        dom = reader.fromString(string)
        return dom

    def getMiniDOMResultFromString(self, string):
        """
        Generate DOM level 1 using minidom (DOM level 1) from stream. Approx. 5 times faster.
        """
        dom = parseString(string)
        return dom

    def setStringParser(self, theParser):
        self._theStringParser = theParser

    def setStreamParser(self, theParser):
        self._theStreamParser = theParser

    def getDOM(self, obj):
        """
        Parse a stream or string with the parser set.
        Returns the DOM.
        """
        if type(obj) == types.StringType:
            self._xml_dom_object = self._theStringParser(obj)
        else:
            self._xml_dom_object = self._theStreamParser(obj)
        
        return self._xml_dom_object


# --------------------------------------------------------------------------

    def getConnection(self):
        """
        Return the connection parameters.
        """
        return self_server, self._database, self._collection

    def setEncoding(self, encoding):
        """
        Specify the encoding to be used for the subsequent requests, e.g. UTF-8.
        """
        self._encoding = encoding

    def getEncoding(self):
        """
        Get the encoding to be used for the subsequent requests.
        Returns the encoding, e.g. 'utf-8'
        """
        return self._encoding

    def request(self, url, cmd = '', query = ''):
        """
        common Tamino request function
        """
        if (cmd == ''):
            return urllib2.urlopen(url)
        
        if (query == ''):
            query = '*'

        if (self._encoding is None):
            params = urllib.urlencode({cmd : query})
        else:
            params = '_encoding=' + self._encoding + '&' + urllib.urlencode({cmd : query})

        if (self._sessionID is not None):
            # transactional request
            tparam = "&_sessionid=%s&_sessionkey=%s" % (self._sessionID , self._sessionKey)
            params = params + tparam

        if (self.isVerbose(1)):
            print "Query: %s" % params

        return urllib2.urlopen(url, params)


    def processResults(self):
        """
        process tamino results with SAX parser
        Returns one Tamino result code. This function is only useful for Tamino X-Machine
        commands that returns one result.
        """

        # self._xml_dom_object = parse(self._fstream)
        #  process results with SAX
        self._saxparser.parse(self._fstream)
        self._sessionID , self._sessionKey = self._results.getSession()
        self._returnValue = self._results.getReturnValue()
        return self._returnValue

# -------------------------------------------------

# Tamino X-Machine Commands

# -------------------------------------------------

    def define(self, stream):
        """
        Define a collection or schema.
        The define command is used to create a new doctype (schema) or a collection, or to modify an existing doctype.
        The collection and / or doctype(s) to be defined or updated have to be expressed in terms of the X-Machine
        schema language. All schema and collection information is stored in the predefined collection ino:collection,
        so it is necessary to add "/ino:collection" to the <URL prefix> in the _define command. 

        DISCLAIMER
        Tamino 2.3 is buggy. It does not return the session ID and session Key when processing _process and _define
        X-Machine commands. It seems that the _define command is not transactional.

        Example:
           <URL prefix>/ino:collection?_define=<schema_definition>

           The input schema definition document defines both the schema and the collection to which it belongs.
           If the collection does not already exist, it is created here automatically. 
        """
    
        if (self._encoding is None):
            params = '_define=*'
        else:
            params = '_encoding=' + self._encoding + '&_define=*'

        if (self._sessionID is not None):
            # transactional request
            tparam = "&_sessionid=%s&_sessionkey=%s" % (self._sessionID , self._sessionKey)
            params = params + tparam

        url = self._url + '?' + params

        if (self.isVerbose(1)):
            print "Request: %s" % url

        http = Resource(url)
        response = http.post( _Define=stream)
        
        if (self.isVerbose(1)):
            print 'HTTP STATUS: ', response.status

        # XXX: Hack, do error handling here. 
        if response.status == 200:
            return response
        else:
            return None  # XXX: Hack, an error occured


    def undefine(self, collectionORschema, magic):
        """
        Delete a collection or schema.
        The undefine command is used to erase an existing doctype (schema) or a collection.
        When you delete a collection or schema, all data belonging to that collection or schema is also deleted.
        In the URL you can either specify the collection name of the collection or schema to be deleted, or you can specify
        the collection name as "ino:collection". Collections whose names start with the characters "ino:" cannot be deleted;
        these are reserved for internal use in Tamino. 

        Use with care! Undefine destroys all data of the collection or schema.
        The magic integer constant must have the numeric value 120664 to perform this operation.
        """
    
        if (magic == 120664):  # the magic is used to prevent acidential method calls.
            self._fstream = self.request(self._url, '_undefine', collectionORschema)
            return self.processResults()
        return -1


    def delete(self, query):
        """
        Delete the documents addressed by the XQL query.
        
        Delete one or more objects in the given collection.
        The delete command is used to delete XML (or Non-XML) objects.
        You can use query language expressions to select the object(s) to be deleted, or refer to an
        object by its object ID.

    Example:
        To delete the XML object of doctype "patient" and with object ID "4711" in the collection "Hospital": 
        <URL prefix>/Hospital?_delete=patient[@ino:id="4711"]

        You can also use the alternative syntax in which the schema name is appended to the collection name: 
        <URL prefix>/Hospital/patient?_delete=patient[@ino:id="4711"]
        """

        self._fstream = self.request(self._url, '_delete', query)
        return self.processResults()


    def deleteByID(self, id, schemaORdoctype=None):
        """
        Delete the documents addressed by the inoID.
        """

        if schemaORdoctype is not None and schemaORdoctype != '':
            query = schemaORdoctype + '[@ino:id="%d"]' % id
        else:
            query = '[@ino:id="%d"]' % id

        self._fstream = self.request(self._url, '_delete', query)
        return self.processResults()



    def process(self, stream, doctype=None):
        """
        Store one or more objects into a collection. Returns the id of the new object(s).
        
        The process command for storing documents into the Tamino database.
        Store one or more objects into a collection. Returns the id of the new object(s).
        Requires a stream object to be stored in the database. Takes an optional argument for the doctype
        of the object to be stored.
        If the object is an XML document (see below for description) then the document is stored
        into the database/collection/doctype if the following conditions apply:
          - the root element of the XML document is equal to the doctype
          - the document structure is valid to the Tamino schema defined in the collection

        Otherwise the document will be stored in the collection ino:etc.

        The process command is used to store a new object into a Tamino database or to modify an existing object.

        DISCLAIMER
        Tamino 2.3 is buggy. Tamino does not return the session ID and session Key when processing _process and _define
        X-Machine commands. We do a bugfix here. It seems that the _process command in Tamino can be undo'd by using
        the old session Key of the previous command. Is anybody else _using_ transactions with Tamino?

        Storing new objects:
        You can insert XML objects and non-XML objects into a Tamino database.
        Any document whose content begins with an XML processing instruction such as
        <?xml version="1.0"?> is considered to be an XML document and will be
        stored in the requested collection and schema. If the non-XML collection was NOT defined,
        all non-XML objects will be stored in the predefined Tamino collection "ino:etc", even if you have
        specified another collection name in the HTTP request.

        If you do not specify a collection name, regardless of whether the
        object is an XML object or not, the object will be stored in the collection
        "ino:etc".

        For an XML object, the root element must be the same as the name of an
        existing schema. The XML object will be stored in the schema with this name. If
        there is no such schema, an error will be returned in the X-Machine response
        document. The specification of the schema name in the URL of the request is
        optional; it is in any case ignored and the schema name is taken from the
        content of the XML object itself.

        Example (raw HTTP Tamino request)
        <URL prefix>/<collection>[/<schema>]?_process=<XML object>
        causes the specified XML object to be stored in the specified
        collection. The schema is the name of the XML object's root node.

        Updating existing objects described as a stream.
        To update an existing object, you need to address it by the object ID
        that was assigned to it when the object was created (the object ID is returned
        in the X-Machine response document when a new object is created in Tamino). To
        update the object, specify the updated data in the HTTP request, and add the
        attribute "ino:id" with the appropriate object ID to the root document.
        """

        if (self._encoding is None):
            params = '_process=*'
        else:
            params = '_encoding=' + self._encoding + '&_process=*'

        if (self._sessionID is not None):
            # transactional request
            tparam = "&_sessionid=%s&_sessionkey=%s" % (self._sessionID , self._sessionKey)
            params = params + tparam

        if (doctype is not None):
            url = self._url + '/' + doctype + '?' + params
        else:
            url = self._url + '?' + params

        if (self.isVerbose(1)):
            print "Request: %s" % url

        http = Resource(url)
        response = http.post( _Process=stream)

        
        if (self.isVerbose(1)):
            print 'HTTP STATUS: ', response.status

        self._fstream = response

        # XXX: Hack, do error handling here. 
        if response.status == 200:
            return response
        else:
            return None  # XXX: Hack, an error occured



    def xql(self, query, start=None, count=None):
        """
        The xql command. Retreive one or more objects and returns a stream of XML or non XML objects.

        Retreive one or more objects. Returns a stream of XML or non XML objects.
        Returns a file stream with the the Tamino result.
        To get the Tamino results use theDOM=getDOM(theStream) in combination with parseTaminoResults(TheDOM)

        The xql command performs a database query to retrieve XML objects.
        X-Machine queries are described in detail in the document Tamino Query Language.

        Example
        To retrieve the list of surnames of all patients in the doctype
        "patient" of the collection "Hospital", use the following:

        <URL prefix>/Hospital/patient?_XQL=patient/name/surname

        The _xql command may be used to retrieve structural information for a doctype or a collection.
        X-Machine answers with a response block containing the internal representation of the schema or the
        collection, expressed in the X-Machine schema language. Note that the structural information of a
        collection or doctype is not stored in the collection or doctype itself but in
        the predefined collection "ino:collection".
        """

        if start is not None and count is not None:
            cmd='_xql(%d,%d)' % (start , count)
        elif start is not None:
            cmd='_xql(%d,)' % start
        elif count is not None:
            cmd='_xql(,%d)' % count
        else:
            cmd='_xql'

        self._fstream = self.request(self._url, cmd, query)
        return self._fstream

    def getContentByID(self, id, schemaORdoctype=None):
        """
        Get content by a given Tamino ID.

        Main use for this function is retrieving binary content or content indexed be a index query (.../@ino:id).
        The tamino id is given as an integer (32-bit). An (optional) doctype or schema can be given as the 2nd argument.
        Returns the content (binary or XML) as stream. Returns a XML message from Tamino if something failed, e.g.
        the content was not found.
        """
        
        if schemaORdoctype is not None and schemaORdoctype != '':
            query = '/' + schemaORdoctype + '/@%d' % id
        else:
            query = '/@%d' % id

        if (self.isVerbose(1)):
            print "Query: %s" % self._url + query
        
        # HTTP GET
        self._fstream = urllib2.urlopen(self._url + query)
        return self._fstream


    def admin(self, query):
        """
        Administration command for administration functions in Tamino

        The admin command offers several administration functions for Tamino.
        """

        self._fstream = self.request(self._url, '_admin', query)
        return self._fstream


    def diagnose(self, query):
        """
        Diagnose command for functionality test.

        The diagnose command is used to test the functionality of the outer X-Machine layers concerned with
        the HTTP handling (the HTTP connection layers).

        <URL prefix>?_diagnose=ping tries to establish an HTTP connection and returns a positive answer in the
        case of success.

        <URL prefix>?_diagnose=echo delivers the HTTP headers seen in the Tamino Server if X-Machine is reachable
        by the web server, otherwise an error message is generated by the web server.

        <URL prefix>?_diagnose=version returns the version of the Tamino Server.

        <URL prefix>?_diagnose=time returns the total amount of time that the server has been active in user
        mode (i.e. executing user requests) since the database server was started, and the total amount of time
        that the server has been active in system (kernel) mode (i.e. executing system calls that result from user
        requests) since the database server was started. The values returned are the total times for all users
        together, not the individual times for each user.

        The X-Machine command "_diagnose" works only with HTTP GET requests but not with HTTP POST requests.

        Since Tamino 2.2., there is a new option for the "_diagnose" command:

        <URL prefix>?_diagnose=apiversion

        This command queries the version number of the Tamino API interfaces.

        Sample of a result document:

        version="1.0" encoding="ISO-8859-1" ?>
        <ino:response xmlns:ino="http://namespaces.softwareag.com/tamino/response2" xmlns:xql="http://metalab.unc.edu/xql/"> <ino:request>
        <ino:diagnose ino:request-type="apiversion"/>
        </ino:request>
        <ino:message>
        <ino:messageline ino:subject="apiVersion">1.1</ino:messageline>
        </ino:message>
        </ino:response>

        The API version number has a major part (the number before the dot) and a minor part (the number after
        the dot). A change in the major part indicates that a change has occurred in the API interface, so that
        it is possible that an application that was written for an older version of the API might not run correctly
        with the new API version. A change in the minor number indicates small enhancements (additional functionality),
        but the API remains compatible with older versions.
        """

        # use HTTP GET here. Diagnose works _only_ with HTTP GET. 
        self._fstream = urllib2.urlopen(self._url + '?' + urllib.urlencode({'_diagnose' : query}))
        return self._fstream


    def parseTaminoResults(self, theDOM):
        """
        Parse the Tamino results and return the most important variables.
        This method is subject to change.

        Parse the given DOM for results for the Tamino DB results.
        Returns:
            Tamino return value
            Tamino message
            cursor
            docID's (list of ino:id)
            list of DOM root elements of the results 
        """
        
        ino = 'http://namespaces.softwareag.com/tamino/response2'
        # transform the Tamino results
        header = theDOM.documentElement.getElementsByTagName("ino:response")
        header = theDOM.getElementsByTagName("ino:response")[0]
        sessionid = header.getAttributeNS(ino, 'sessionid')
        
        if sessionid != '':
            self._sessionID = sessionid
            sessionkey = header.getAttributeNS(ino, 'sessionkey')
            if sessionkey != '':
                self._sessionKey = sessionkey
        else:
            # Tamino is buggy. It does not return the session ID and session Key when processing _process and _define
            # X-Machine commands. We do a bugfix here. Another bug: the _process command can be undo'd by using
            # the old session Key of the previous command. Is anybody out there using transactions with Tamino?
            #self._sessionID = None
            #self._sessionKey = None
            pass

        returnvalue = 0
        messageline = ''
        cursor = 0
        docIDs = []
        resultDOMs = []
        first=1

        # parse Tamino results
        # Code change 2.1.2002: Beware of another Tamino bug that does not permit to store
        # documents with length 0 (zero bytes length). The tamino return code in this case is
        # 8554 and the document is not created.

        for node in header.childNodes:
            # HACKHACK: This is a hack. The parsing of the Tamino return values is badly hardcoded here.
            if node.nodeName == 'ino:message':
                if not first:
                    continue

                first=0
                attr = node.getAttributeNS(ino, 'returnvalue')
                if attr != '':
                    returnvalue = int(attr)

                subnode = node.getElementsByTagNameNS(ino, 'messageline')
                if len(subnode) > 0:  # HACK: get only the first messageline
                    subnode = subnode[0].firstChild
                    if subnode is not None and subnode.nodeType == subnode.TEXT_NODE:
                        messageline = subnode.nodeValue
            elif node.nodeName == 'ino:cursor':
                attr = node.getAttributeNS(ino, 'count')
                if attr != '':
                    cursor = int(attr)
            elif node.nodeName == 'xql:result':
                # XQL command: get ino:id and subDOM's of the documents returned
                if node.hasChildNodes:
                    resultDOMs = node.childNodes
                    # print resultDOMs
                    for subnode in resultDOMs:
                        attr = subnode.getAttributeNS(ino, 'id')
                        if attr != '':
                            docIDs.append(int(attr))
            elif node.nodeName == 'ino:object':
                # PROCESS command: get ino:id of the documents processed
                attr = node.getAttributeNS(ino, 'id')
                if attr != '':
                    docIDs.append(int(attr))
        
        if (self.isVerbose(1)):
            print 'Returnvalue: ', returnvalue
            print 'Messageline: ', messageline
            print 'SESSIONID: %s' % self._sessionID
            print 'SESSIONKEY: %s' % self._sessionKey
            print 'Cursor: %d' % cursor
            print "Document ID's", docIDs
            print 'Results found: %d' % len(resultDOMs)
            print 'Result root elements: %d' % len(resultDOMs)

        # return all the results
        return returnvalue, messageline, docIDs, resultDOMs, cursor



    def applyXSLT(self, theDOMorXMLString, theStyleSheet):
        """
        A sample implementation how to apply a style sheet to the DOM. Be aware that the 4Suite XSLT is
        not the fastest stylesheet processor. It seems that 4Suite XSLT requires the 4DOM representation of the DOM.
        XSLT processing from a document string is much faster then processing the XML document as a DOM. 
        4DOM is about 5 times slower then miniDOM.

        Example:
            stylesheet = open('myXSLT.xsl')
            print applyXSLT(myDOM, stylesheet)

        NOTE: This method is for demonstation purposes only. Do not call this method from your
              application (because taminotalk is a database interface).
        """

        processor=Processor()
        processor.appendStylesheetStream(theStyleSheet)

        if type(theDOMorXMLString) == types.StringType:
            return processor.runString(theDOMorXMLString)
        else:
            return processor.runNode(theDOMorXMLString)

# ----------------------------------------------------------------------------------
# Commands for transaction control.
#

    def connect(self, isolation=None, lockwait=None):
        """
        Start a transactional database session. Not required for non transactional sessions.

        The _isolation Parameter

         The _isolation parameter specifies in what way two or more applications can access the same data simultaneously. If one session is currently accessing data from the database and a second
         session attempts to access or modify the same data, the setting of the _isolation parameter for each of the sessions determines whether access is possible for the second session. 

         The _isolation parameter can also be specified in a non-session context on a single command. This controls the behaviour of the command if it accesses data that is currently under the control of
         a transaction from another application or another command using the _isolation parameter in a non-session context. 

         The _isolation parameter can have the following values:

               _isolation=unprotected
               _isolation=shared
               _isolation=protected

      The _lockwait Parameter

         The _lockwait parameter specifies which action to take if data is currently not accessible because another transaction has used the _isolation parameter to restrict access to the data. This
         parameter can have the following values: 

          _lockwait=yes
                               If data is locked by another transaction, wait until the data is unlocked by the other
                               transaction and then continue processing. 
          _lockwait=no
                               If data is locked by another transaction, terminate the session immediately and return
                               an error message in the <ino:message> section of the response body. 

        """

        cmd='_connect=*'
        if isolation is not None:
            cmd = cmd + '&_isolation=' + isolation
        if lockwait is not None:
            cmd = cmd + '&_lockwait=' + lockwait

        self._fstream = urllib2.urlopen(self._url, cmd)
        return self.processResults()

    def disconnect(self):
        """
        Terminate the database session, commits all changes (use for transactional access only).
        """

        if (self._sessionID is None):
            return -1  # return error code
        self._fstream = self.request(self._url, '_disconnect')
        ret = self.processResults()
        
        # clear sessions
        self._sessionID = None
        self._sessionKey = None
        return ret


    def commit(self):
        """
        Commit a transaction (use for transactional access only)
        """

        if (self._sessionID is None):
            return -1  # return error code
        self._fstream = self.request(self._url, '_commit')
        return self.processResults()


    def rollback(self):
        """
        Rollback a transaction (use for transactional access only).
        """
        
        if (self._sessionID is None):
            return -1  # return error code
        self._fstream = self.request(self._url, '_rollback')
        return self.processResults()


# -------------------------------------------------------------------------------
# test code goes here
# -------------------------------------------------------------------------------
# This is some test code for the Shakespeare data base at struktur AG.
# The test code is *not* for general purpose use.


def test1(args = None):
# server configuration goes here
# Put a document in the database
    server = '10.1.1.11'
    database = 'sp'
    collection = 'movies'
    doctype = 'mpeg'

# connect tor the database server
    t = TaminoTalk()
    t.connectTo(server, database, collection)
    t.setVerbose(1)

    print '\nADMIN'

    ret = t.admin('ino:CheckCompatible ("V2.2")')
# create a XML DOM from database stream.
    theDOM = t.getDOM(ret)
    xml.dom.ext.PrettyPrint(theDOM)
    values = t.parseTaminoResults(theDOM)
    # print out the results
    print "Returnvalues: ", values

    print '\nDIAGNOSE (version)'

    ret = t.diagnose('version')
# create a XML DOM from database stream.
    theDOM = t.getDOM(ret)
    xml.dom.ext.PrettyPrint(theDOM)
    values = t.parseTaminoResults(theDOM)
    # print out the results
    print "Returnvalues: ", values


    print '\nPROCESS'

# open a file.
    fp = open('a.a', 'r')
    print '------------------------------------------------------'

# start transaction. connect(), commit() and terminate() is ONLY required for transactional data processing.
    print 'Connect'
    ret = t.connect()
    print "Returnvalue: %s" % ret

# put the document into the given database/collection/doctype
    response = t.process(fp, doctype)
    fp.close()

    # parse response
    if response is not None:
        theDOM = t.getDOM(response)
        response.close()
        xml.dom.ext.PrettyPrint(theDOM)
        values = t.parseTaminoResults(theDOM)
        # print out the results
        print "Returnvalues: ", values
    else:
        print "ERROR: Process command failed"

# roolback the transaction
    print 'Rollback transaction'
    ret = t.rollback()
    print "Returnvalue: %s" % ret

# disconnect
    print 'Disconnect'
    ret = t.disconnect()
    print "Returnvalue: %s" % ret



#   print '\nDELETE'

# put the document into the given database/collection/doctype
#   ret = t.deleteByID(50, doctype)
#   ret = t.deleteByID(51, doctype)
    
#   print "Returnvalue: %s" % ret


# ---------------------------------------------------------------------------------------------
def test2(args = None):
# server configuration goes here
# Put a document in the database
    server = '10.1.1.11'
    database = 'sp'
    collection = 'movies'
    doctype = 'mpeg'

# connect tor the database server
    t = TaminoTalk()
    t.connectTo(server, database, collection)
    t.setVerbose(1)

    print '\nPROCESS'

    import StringIO

    usc = u'dies ist ein blder Test mit {}|<>,.;:_#'
    str = usc.encode('utf-8')
    s = StringIO.StringIO(str)

#   s = StringIO.StringIO(u'12345')

    s.seek(0)
    print s.read()
    s.seek(0)

    print '------------------------------------------------------'


# start transaction. connect(), commit() and terminate() is ONLY required for transactional data processing.
    print 'Connect'
    ret = t.connect()
    print "Returnvalue: %s" % ret

# put the document into the given database/collection/doctype
    response = t.process(s, doctype)
    s.close()

    # parse response
    if response is not None:
        theDOM = t.getDOM(response)
        response.close()
        xml.dom.ext.PrettyPrint(theDOM)
        values = t.parseTaminoResults(theDOM)
        # print out the results
        print "Returnvalues: ", values
    else:
        print "ERROR: Process command failed"

# roolback the transaction
#   print 'Rollback transaction'
#   ret = t.rollback()
#   print "Returnvalue: %s" % ret

# disconnect
    print 'Disconnect'
    ret = t.disconnect()
    print "Returnvalue: %s" % ret


# ---------------------------------------------------------------------------------------------



def test(args = None):
# server configuration goes here
    server = '10.1.1.11'
    database = 'sp'
    collection = 'publish'


# XQL queries
    queries = { 'all IDs' : 'PLAY/@ino:id',
              'all documents' : 'PLAY',
              'midsummer docs' : "PLAY/[TITLE~='Midsummer']",
              'personae lysander' : "PLAY[//PERSONA~='LYSANDER']"
              }


# connect tor the database server
    t = TaminoTalk()
    t.connectTo(server, database, collection)
    t.setVerbose(1)
    t.setEncoding('utf-8')

# start transaction. connect(), commit() and terminate() is ONLY required for transactional data processing.
    print 'Connect'
    ret = t.connect()
    print "Returnvalue: %s" % ret

# query the database
    print '\nXQL'
# query the database. xql() returns the raw database stream.
    ret = t.xql('PLAY', 1, 1)

# create a XML DOM from database stream.
    theDOM = t.getDOM(ret)

# parse the results of the query.
# results are in the form: (returnvalue, messageline, docIDs, resultDOMs, cursor)
    values = t.parseTaminoResults(theDOM)
    print "Returnvalues: ", values 

# commit the tansaction and terminate transaction. commit() is not required here.
# commit is called for testing purposes only.
    print 'Commit'
    ret = t.commit()
    print "Returnvalue: %s" % ret

# terminate the database session. Not further transactions are allowed from this point.
# To establish a new connections, call connect() again.
    print 'Disconnect'
    ret = t.disconnect()
    print "Returnvalue: %s" % ret

# test the data encoding to default encoding.
    t.setEncoding(None)

    print '\nXQL'
# query the database. xql() returns the raw database stream.
    ret = t.xql('PLAY/@ino:id', 3, 2)

# create a XML DOM from database stream.
    theDOM = t.getDOM(ret)

# parse the results of the query.
# results are in the form: (returnvalue, messageline, docIDs, resultDOMs, cursor)
    values = t.parseTaminoResults(theDOM)
    print "Returnvalues: ", values 


    print '\nXQL'
# query the database. xql() returns a stream.
    ret = t.xql('PLAY/@ino:id', 3, 2)
# read a stream in a string 
    str = ret.read()

# create a XML DOM from result string.
    theDOM = t.getDOM(str)

# parse the results of the query.
# results are in the form: (returnvalue, messageline, docIDs, resultDOMs, cursor)
    values = t.parseTaminoResults(theDOM)
    print "Returnvalues: ", values 

# query the database by a given content ID
    print '\nGet content by ID:'
    stream = t.getContentByID(1, 'PLAY')

# to print the stream
#   print stream.read()

# query the database. xql() returns a stream.
    print '\nXQL'
    ret = t.xql('PLAY', 1, 1)

# read a stream in a string 
    str = ret.read()

# create a XML DOM from result string.
    theDOM = t.getDOM(str)

# parse the results of the query.
# results are in the form: (returnvalue, messageline, docIDs, resultDOMs, cursor)
    values = t.parseTaminoResults(theDOM)
    print "Returnvalues: ", values 

# apply a XSL stylesheet to the string result of the query
    print "---------------------------"
    try:
        print t.applyXSLT(str, open('title.xsl'))
    except:
        print "XSLT test skipped."
    print "---------------------------"


# Put a document in the database
#   server = '10.1.1.11'
#   database = 'sp'
    collection = 'movies'
    doctype = 'mpeg'

# connect tor the database server
    t = TaminoTalk()
    t.connectTo(server, database, collection)
    t.setVerbose(1)

    print '\nPROCESS'

# open a file.
    fp = open('taminotalk.py', 'r')
    print '------------------------------------------------------'

# start transaction. connect(), commit() and terminate() is ONLY required for transactional data processing.
    print 'Connect'
    ret = t.connect()
    print "Returnvalue: %s" % ret

# put the document into the given database/collection/doctype
    response = t.process(fp, doctype)
    fp.close()

    # parse response
    if response is not None:
        theDOM = t.getDOM(response)
        response.close()
        values = t.parseTaminoResults(theDOM)
        # print out the results
        print "Returnvalues: ", values
    else:
        print "ERROR: Process command failed"

# roolback the transaction
    print 'Rollback transaction'
    ret = t.rollback()
    print "Returnvalue: %s" % ret

# disconnect
    print 'Disconnect'
    ret = t.disconnect()
    print "Returnvalue: %s" % ret

# define a new collection
    t = TaminoTalk()
    t.connectTo(server, database)
    t.setVerbose(1)

    print '\nDEFINE'

# open a file with a schema.
    fp = open('Schemas/binary_ts.xml', 'r')
    print '------------------------------------------------------'

# start transaction. connect(), commit() and terminate() is ONLY required for transactional data processing.
    print 'Connect'
    ret = t.connect()
    print "Returnvalue: %s" % ret

# define the schema or doctype
    response = t.define(fp)
    fp.close()
    
    # parse response
    if response is not None:
        theDOM = t.getDOM(response)
        response.close()
        values = t.parseTaminoResults(theDOM)
        # print out the results
        print "Returnvalues: ", values
        # optional print the XML document with nice formatting
        # xml.dom.ext.PrettyPrint(theDOM)
    else:
        print "ERROR: Process command failed"


# roolback the transaction
    print 'Rollback transaction'
    ret = t.rollback()
    print "Returnvalue: %s" % ret

# disconnect
    print 'Disconnect'
    ret = t.disconnect()
    print "Returnvalue: %s" % ret

# the following lines demonstrate a bug in Tamino 2.3.x: Defines are *not* transactional and therefore cannot be undo'd
    print '\nUNDEFINE'
    ret=t.undefine('publish2', 120664)
    print "Returnvalue: %s" % ret


# -------------------------------------------------------------------------------
# main class
#

if __name__ == '__main__':
    test()
#        test1()
